import sys
import os
modules_path = '../../../modules/'
if os.path.exists(modules_path):
    sys.path.insert(1, modules_path)

import json
from datetime import datetime, date
from db.repository import Repository
from db.entities import ClusterSet, Cluster, Layer, TimeSlice
from typing import Tuple, Dict, Any, List

TimeSliceKey = Tuple[int, int]

# TODO extract information about time features (maybe from table mapping)
TIME_PROPERTY_NAMES = ['timestamp']

repo = Repository()


def convert_to_time_slice_key(timestamp: str) -> TimeSliceKey:
    '''Returns the tuple (year, week_of_year) from a timestamp. This is used as the key for the slicing.'''
    # time = datetime.utcfromtimestamp(float(timestamp[0:10]))
    # time = datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
    time = datetime.utcfromtimestamp(float(timestamp))
    (y, w, _) = time.isocalendar()
    return (y, w)


def split_clusterset_by_time(layer: Layer, clusters: List[Cluster]) -> Dict[TimeSliceKey, TimeSlice]:
    '''
    Distributes all nodes in clusters of a single layer into individual time slices based on their timestamps. 
    If a node spans over multiple slices it will be added to all of them.
    Information about clusters and the nodes in the clusters will not be changed.

    :params clusters: The clusters whichs nodes are split
    :returns: A dict of time slices where the key is the time info and value is the information about the time slice
    '''

    time_slices: Dict[Any, TimeSlice] = {} 
    all_cluster_labels = [cluster.cluster_label for cluster in clusters if cluster.cluster_label != -1]

    for cluster in clusters:
        if cluster.cluster_label == -1:
            print("Noise cluster was ignored.")
            continue

        for node in cluster.nodes:

            # retrieve times where the node is located inas (year, week), based on the defined timestamp fields in the schema
            time_keys = set()
            for time_property in TIME_PROPERTY_NAMES:
                if time_property in node:
                    time_keys.add(convert_to_time_slice_key(str(node[time_property])))  

            for time_key in time_keys:
                if time_key not in time_slices:
                    time_slices[time_key] = TimeSlice(time_key, layer.use_case, layer.use_case_table, layer.layer_name)
                    time_slices[time_key].init_all_clusters(all_cluster_labels)

                time_slices[time_key].add_node_to_cluster(cluster.cluster_label, node)

    return time_slices
        

def get_layers() -> List[Layer]:
    return repo.get_layers()

def get_clusters_for_layer(use_case, use_case_table, layer_name)-> List[Cluster]:
    # return repo.get_clusters_for_layer(use_case, use_case_table, layer_name)
    json_path = f'_predictions/clusters/{layer_name}.json'

    if os.path.exists(json_path):
        with open(json_path, 'r') as file:
            return [Cluster(cluster_dict=e, from_db=False) for e in json.loads(file.read())]

    return []

def get_layer_nodes(use_case, use_case_table, layer_name)-> List[dict]:
    # return repo.get_layer_nodes(use_case, use_case_table, layer_name)
    return []

def add_time_slice(timeslice):
    try:
        # repo.add_time_slice(timeslice)  
        pass
    except:
        print(f"Error while storing time slice in db for {timeslice.layer_name}")

    try:
        json_path = f'_predictions/timeslices/{timeslice.layer_name}/{timeslice.time}.json'.replace(', ', '_').replace('(', '').replace(')', '')
        if not os.path.exists(os.path.dirname(json_path)):
            os.makedirs(os.path.dirname(json_path))

        with open(json_path, 'w') as file:
            file.write(json.dumps(timeslice.to_serializable_dict(for_db=False)))
    except Exception as e:
        print(f"Error while writing json for {timeslice.layer_name}: {e}")

def run_time_slicing(selected_use_cases: List[str] = None, selected_use_case_tables: List[str] = None, selected_layer_names: List[str] = None):
    layers = get_layers()
    for layer in layers:
        layer_name = layer.layer_name
        use_case = layer.use_case
        use_case_table = layer.use_case_table

        # skip layers not in the params
        if selected_use_cases is not None and use_case not in selected_use_cases \
            or selected_use_case_tables is not None and use_case_table not in selected_use_case_tables \
            or selected_layer_names is not None and layer_name not in selected_layer_names:
            continue

        print(f"Working on {use_case}//{use_case_table}//{layer_name}.")

        clusters_for_layer = get_clusters_for_layer(use_case, use_case_table, layer_name)
        # if no clusters were generated use one large cluster instead of skipping the layer
        if clusters_for_layer is None or len(clusters_for_layer) == 0:
            nodes = get_layer_nodes(use_case, use_case_table, layer_name)
            if nodes is None or len(nodes) == 0:
                print("Skipping, because there are no clusters and no nodes for the layer.")
                continue
            clusters_for_layer = [Cluster(use_case, use_case_table, layer_name, -1, nodes,  'noise')]

        time_slices = split_clusterset_by_time(layer, clusters_for_layer)

        for k,v in time_slices.items():
            add_time_slice(v)


if __name__ == "__main__":
    # repo.remove_all_time_slices()
    run_time_slicing(selected_use_cases=['community-prediction-taxi'])